/* acl.c - access control lists

   Copyright (C) 2002 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   Written by Paul Eggert.  */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdlib.h>
#include <sys/stat.h>
#ifndef S_ISLNK
# define S_ISLNK(Mode) 0
#endif

#include "error.h"
#include "quote.h"
#include "acl.h"

#ifdef HAVE_ACL_LIBACL_H
# include <acl/libacl.h>
#endif

#include <errno.h>
#ifndef ENOSYS
# define ENOSYS (-1)
#endif
#ifndef ENOTSUP
# define ENOTSUP (-1)
#endif

#if ENABLE_NLS
# include <libintl.h>
# define _(Text) gettext (Text)
#else
# define _(Text) Text
#endif

#ifndef MIN_ACL_ENTRIES
# define MIN_ACL_ENTRIES 4
#endif

/* Return 1 if PATH has a nontrivial access control list, 0 if not,
   and -1 (setting errno) if an error is encountered.  */

int
file_has_acl (char const *path, struct stat const *pathstat)
{
#if USE_ACL && HAVE_ACL && defined GETACLCNT
  /* This implementation should work on recent-enough versions of HP-UX,
     Solaris, and Unixware.  */

  if (! S_ISLNK (pathstat->st_mode))
    {
      int n = acl (path, GETACLCNT, 0, NULL);
      return n < 0 ? (errno == ENOSYS ? 0 : -1) : (MIN_ACL_ENTRIES < n);
    }
#elif USE_ACL && HAVE_ACL_EXTENDED_FILE
  /* Linux specific version.  */

  if (! S_ISLNK (pathstat->st_mode))
    {
      int ret = acl_extended_file (path);
      if (ret < 0)
	return (errno == ENOSYS || errno == ENOTSUP) ? 0 : -1;
      return ret;
    }
#endif

  /* FIXME: Add support for POSIX 1003.1e (draft 17 -- abandoned), AIX, Irix,
     and Tru64.  Please see Samba's source/lib/sysacls.c file for fix-related
     ideas.  */

  return 0;
}

/* Copy the access control list of src_path to dst_path. Fall back to
   src_st.st_mode if access control lists are not supported for either
   file.  */
int
copy_acl (char const *src_path, char const *dst_path, mode_t mode)
{
#if USE_ACL && HAVE_ACL_GET_FILE && HAVE_ACL_SET_FILE && \
    HAVE_ACL_FREE && HAVE_ACL_ENTRIES
  /* Linux specific. Will work on all POSIX 1003.1e draft 17 (abandoned)
     systems if the Linux specific acl_entries() function is substituted.  */

  acl_t acl = acl_get_file (src_path, ACL_TYPE_ACCESS);
  if (acl == NULL)
    {
      if (errno == ENOSYS || errno == ENOTSUP)
	return set_acl (dst_path, mode);
      else
        {
	  error (0, errno, "%s", quote (src_path));
	  return -1;
	}
    }

  if (acl_set_file (dst_path, ACL_TYPE_ACCESS, acl))
    {
      int saved_errno = errno;

      if (errno == ENOSYS || errno == ENOTSUP)
        {
	  int n = acl_entries (acl);

	  acl_free (acl);
	  if (n == 3)
	    {
	      if (chmod (dst_path, mode))
		saved_errno = errno;
	      else
		return 0;
	    }
	  else
	    chmod (dst_path, mode);
	}
      else
	{
	  acl_free (acl);
	  chmod (dst_path, mode);
	}
      error (0, saved_errno, _("preserving permissions for %s"),
	     quote (dst_path));
      return -1;
    }
  else
    acl_free (acl);

  if (mode & (S_ISUID | S_ISGID | S_ISVTX))
    {
      /* We did not call chmod so far, so the special bits have not yet
         been set.  */

      if (chmod (dst_path, mode))
	{
	  error (0, errno, _("preserving permissions for %s"),
	  	 quote (dst_path));
	  return -1;
	}
    }

  if (S_ISDIR (mode))
    {
      acl = acl_get_file (src_path, ACL_TYPE_DEFAULT);
      if (acl == NULL)
	{
	  error (0, errno, "%s", quote (src_path));
	  return -1;
	}

      if (acl_set_file (dst_path, ACL_TYPE_DEFAULT, acl))
	{
	  error (0, errno, _("preserving permissions for %s"),
		 quote (dst_path));
	  acl_free(acl);
	  return -1;
	}
      else
        acl_free(acl);
    }
  return 0;
#else
  int ret = chmod (dst_path, mode);
  if (ret)
    error (0, errno, _("preserving permissions for %s"), quote (dst_path));
  return ret;
#endif
}

/* Set the access control list of path to the permissions defined by mode.  */
int
set_acl (char const *path, mode_t mode)
{
#if USE_ACL && HAVE_ACL_FROM_TEXT && HAVE_ACL_SET_FILE && HAVE_ACL_FREE && \
    HAVE_ACL_DELETE_DEF_FILE
  /* POSIX 1003.1e draft 17 (abandoned) specific version.  */

  char acl_text[] = "u::---,g::---,o::---";
  acl_t acl;

  if (mode & S_IRUSR) acl_text[ 3] = 'r';
  if (mode & S_IWUSR) acl_text[ 4] = 'w';
  if (mode & S_IXUSR) acl_text[ 5] = 'x';
  if (mode & S_IRGRP) acl_text[10] = 'r';
  if (mode & S_IWGRP) acl_text[11] = 'w';
  if (mode & S_IXGRP) acl_text[12] = 'x';
  if (mode & S_IROTH) acl_text[17] = 'r';
  if (mode & S_IWOTH) acl_text[18] = 'w';
  if (mode & S_IXOTH) acl_text[19] = 'x';

  acl = acl_from_text(acl_text);
  if (!acl)
    {
      error (0, errno, "%s", quote (path));
      return -1;
    }

  if (acl_set_file(path, ACL_TYPE_ACCESS, acl))
    {
      int saved_errno = errno;
      acl_free (acl);

      if (errno == ENOTSUP || errno == ENOSYS)
	{
	  if (chmod (path, mode))
	    saved_errno = errno;
	  else
	    return 0;
	}
      error (0, saved_errno, _("setting permissions for %s"), quote (path));
      return -1;
    }
  else
    acl_free (acl);

  if (mode & (S_ISUID | S_ISGID | S_ISVTX))
    {
      /* We did not call chmod so far, so the special bits have not yet
         been set.  */

      if (chmod (path, mode))
	{
	  error (0, errno, _("preserving permissions for %s"),
	  	 quote (path));
	  return -1;
	}
    }

  if (S_ISDIR (mode) && acl_delete_def_file (path))
    {
      error (0, errno, _("setting permissions for %s"), quote (path));
      return -1;
    }
  return 0;
#else
  int ret = chmod (path, mode);
  if (ret)
    error (0, errno, _("setting permissions for %s"), quote (path));
  return ret;
#endif
}
