/* GNU's read utmp module.
   Copyright (C) 1992-2001 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/* Written by jla; revised by djm */

#include <config.h>

#include <stdio.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <stdbool.h>
#if defined(STDC_HEADERS) || defined(HAVE_STRING_H)
# include <string.h>
#else
# include <strings.h>
#endif /* STDC_HEADERS || HAVE_STRING_H */

#include "readutmp.h"
#include "unlocked-io.h"

char *xmalloc ();
char *realloc ();
char *xrealloc();

#ifndef SIZE_MAX
# define SIZE_MAX ((size_t) -1)
#endif

/* Copy UT->ut_name into storage obtained from malloc.  Then remove any
   trailing spaces from the copy, NUL terminate it, and return the copy.  */

char *
extract_trimmed_name (const STRUCT_UTMP *ut)
{
  char *p, *trimmed_name;

  trimmed_name = xmalloc (sizeof (UT_USER (ut)) + 1);
  strncpy (trimmed_name, UT_USER (ut), sizeof (UT_USER (ut)));
  /* Append a trailing space character.  Some systems pad names shorter than
     the maximum with spaces, others pad with NULs.  Remove any spaces.  */
  trimmed_name[sizeof (UT_USER (ut))] = ' ';
  p = strchr (trimmed_name, ' ');
  if (p != NULL)
    *p = '\0';
  return trimmed_name;
}

/* Is the utmp entry U desired by the user who asked for OPTIONS?  */

static inline bool
desirable_utmp_entry (STRUCT_UTMP const *u, int options)
{
#define UT_TYPE_UNDEF 255

#if HAVE_STRUCT_XTMP_UT_TYPE
# define UT_TYPE(U) ((U)->ut_type)
#else
# define UT_TYPE(U) UT_TYPE_UNDEF
#endif

#define IS_USER_PROCESS(U)                      \
  (UT_USER (U)[0]				\
   && (UT_TYPE (U) == USER_PROCESS		\
       || (UT_TYPE (U) == UT_TYPE_UNDEF		\
           && UT_TIME_MEMBER (U) != 0)))

  return ! (options & READ_UTMP_CHECK_PIDS
	    && IS_USER_PROCESS (u)
	    && (UT_PID (u) <= 0
		|| (kill (UT_PID (u), 0) < 0 && errno == ESRCH)));
}

/* Read the utmp entries corresponding to file FILENAME into freshly-
   malloc'd storage, set *UTMP_BUF to that pointer, set *N_ENTRIES to
   the number of entries, and return zero.  If there is any error,
   return -1, setting errno, and don't modify the parameters.
   If OPTIONS & READ_UTMP_CHECK_PIDS is nonzero, omit entries whose
   process-IDs do not currently exist.  */

#ifdef UTMP_NAME_FUNCTION

int
read_utmp (const char *filename, int *n_entries, STRUCT_UTMP **utmp_buf,
	   int options)
{
  int n_read = 0;
  STRUCT_UTMP *u;
  STRUCT_UTMP *utmp = NULL;

  /* Ignore the return value for now.
     Solaris' utmpname returns 1 upon success -- which is contrary
     to what the GNU libc version does.  In addition, older GNU libc
     versions are actually void.   */
  UTMP_NAME_FUNCTION (filename);

  SET_UTMP_ENT ();

  while ((u = GET_UTMP_ENT ()) != NULL)
    if (desirable_utmp_entry (u, options))
      {
	++n_read;
	utmp = (STRUCT_UTMP *) realloc (utmp, n_read * sizeof (STRUCT_UTMP));
	if (utmp == NULL)
	  return 1;
	utmp[n_read - 1] = *u;
      }

  END_UTMP_ENT ();

  *n_entries = n_read;
  *utmp_buf = utmp;

  return 0;
}

#else
#define UTMP_READ_INCR 128

int
read_utmp (const char *filename, int *n_entries, STRUCT_UTMP **utmp_buf)
{
    int n_read;
    STRUCT_UTMP *uptr;

    utmpname(filename);
    setutent();
    n_read = 0;
    utmp_contents = (STRUCT_UTMP *) xmalloc(UTMP_READ_INCR*sizeof(STRUCT_UTMP));
    while ((uptr=getutent())) {
	if (NULL == utmp_contents)
	    break;  /* we do not have more memory here */
	if (!desirable_utmp_entry (uptr, options))
	    continue;
	memcpy(&utmp_contents[n_read], uptr, sizeof(STRUCT_UTMP));
	n_read++;
	if (0 == ( n_read % UTMP_READ_INCR ))
	    utmp_contents = (STRUCT_UTMP *)
		xrealloc(utmp_contents,
			 (n_read + UTMP_READ_INCR) * sizeof(STRUCT_UTMP));
    }
   
    endutent();
    return n_read;
}

#endif
